/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Brain
 * Description: AI routines and the interface to the PCs "brain" (player control)
 */

#define Uses_Ui
#define Uses_Command
#define Uses_Character
#define Uses_Stats
#define Uses_Movement
#define Uses_Util
#define Uses_Pathfinder
#define Uses_Actions
#define Uses_Area
#define Uses_Sector
#define Uses_Object
#define Uses_Inventory
#define Uses_Faction
#define Uses_Combat
#define Uses_Perception
#define Uses_Dice
#define Uses_Psychic
#define Uses_Career
#include "mheader.h"

#include "brain.h"


/*
 * number of AI states
 */
#define MAX_AI_STATES				19


/*
 * AI decision probabilities (1% - 100% chance)
 */
#define AI_SWITCH_TO_GUARD_PROBABILITY		1
#define AI_SWITCH_TO_PATROL_PROBABILITY		2

#define AI_PSYKER_DEFENSIVE_ACTION_PROBABILITY	25


/*
 * the AI will not use concentration unless the difficulty modifier
 * of the power it tries to evoke is higher than or equal to this
 */
#define AI_MIN_DIFFICULTY_TO_CONCENTRATE	20


/*
 * the maxium level of concentration the AI will use
 * while not following the "Get ready!" tactic
 */
#define AI_NORMAL_CONCENTRATION			10



static void	ai_execute(CHARACTER *);

static AI_STATE	ai_state_wait(CHARACTER *);
static AI_STATE	ai_state_guard(CHARACTER *);
static AI_STATE	ai_state_patrol(CHARACTER *);
static AI_STATE	ai_state_follow_pcc(CHARACTER *);

static AI_STATE	ai_state_combat(CHARACTER *);
static AI_STATE	ai_state_combat_run_towards_target(CHARACTER *);

static AI_STATE	ai_state_psyker_combat(CHARACTER *);
static AI_STATE	ai_state_psyker_defensive(CHARACTER *);
static AI_STATE	ai_state_psyker_offensive(CHARACTER *);
static AI_STATE	ai_state_psyker_evocation(CHARACTER *);

static AI_STATE	ai_state_attack(CHARACTER *);

static AI_STATE	ai_state_attack_target_close(CHARACTER *);
static AI_STATE	ai_state_attack_target_distant(CHARACTER *);

static AI_STATE	ai_state_attack_strike(CHARACTER *);
static AI_STATE	ai_state_attack_shoot(CHARACTER *);

static AI_STATE	ai_state_recover(CHARACTER *);

static AI_STATE	ai_state_flee(CHARACTER *);

static AI_STATE	ai_state_player_control(CHARACTER *);

static void	ai_update_target_distance(CHARACTER *);

static void	ai_fix_self(CHARACTER *);

static bool	ai_use_drug(CHARACTER *, const char *);

static bool	ai_close_combat_focused(const CHARACTER *);

static bool	ai_psy_power_sensible_choice(PSY_POWER,
			const AREA_POINT *
		);

static bool	ai_close_to_pcc(const CHARACTER *);

static void	ai_erase_target(CHARACTER *);

static bool	ai_target_reached(const CHARACTER *);

static bool	ai_has_target(const CHARACTER *);

static void	ai_get_ready_action(CHARACTER *);
static void	ai_get_ready(CHARACTER *);
static void	ai_ready_weapon(CHARACTER *);
static void	ai_get_ready_psyker(CHARACTER *);

static void	ai_update_awareness_data(CHARACTER *);



/*
 * AI state functions
 */
static AI_STATE (*AiState[MAX_AI_STATES])(CHARACTER *) = {

	ai_state_wait,
	ai_state_guard,
	ai_state_patrol,
	ai_state_follow_pcc,

	ai_state_combat,
	ai_state_combat_run_towards_target,

	ai_state_psyker_combat,
	ai_state_psyker_defensive,
	ai_state_psyker_offensive,
	ai_state_psyker_evocation,

	ai_state_attack,

	ai_state_attack_target_close,
	ai_state_attack_target_distant,

	ai_state_attack_strike,
	ai_state_attack_shoot,

	ai_state_recover,

	ai_state_flee,

	ai_state_player_control
};



/*
 * AI tactic names
 */
const char * AiTacticName[MAX_AI_TACTICS] = {

	"Plan A!",
	"Get ready!",
	"Stay close!",
	"Hold position!"

};




/*
 * activates a character's "brain" - makes the character act
 */
void brain(CHARACTER *character)
{

	ai_update_awareness_data(character);
	
	if (ai_has_target(character)) {

		ai_update_target_distance(character);
	}
	
	if (character->ai.state != AI_STATE_PLAYER_CONTROL) {

		ai_fix_self(character);
	}

	while (!character->action_spent) {

		ai_execute(character);
	}
}





/*
 * sets the target of the AI
 */
void ai_set_target(CHARACTER *character, const AREA_POINT *target_point)
{

	character->ai.target_point = *target_point;
	
	character->ai.target_distance = area_distance(
		&character->location, target_point
	);
}



/*
 * sets the tactic of the AI
 */
void ai_set_tactic(CHARACTER *character, AI_TACTIC tactic)
{

	character->ai.tactic = tactic;

	if (tactic == TACTIC_GET_READY ||
		tactic == TACTIC_HOLD_POSITION) {

		character->ai.state = AI_STATE_GUARD;

	} else {

		character->ai.state = AI_STATE_FOLLOW_PCC;
	}

	character->ai.default_state = character->ai.state;
}



/*
 * returns the name of the passed AI tactic
 */
const char * ai_tactic_name(AI_TACTIC tactic)
{

	return AiTacticName[tactic];
}



/*
 * executes the AI
 */
static void ai_execute(CHARACTER *character)
{

	character->ai.state =
		(*AiState[character->ai.state])(character);
}



/*
 * AI state: wait
 */
static AI_STATE ai_state_wait(CHARACTER *character)
{

	if (enemies_noticed(character)) {

		return AI_STATE_COMBAT;
	}
	
	ai_get_ready_action(character);

	return AI_STATE_WAIT;
}



/*
 * AI state: guard
 */
static AI_STATE ai_state_guard(CHARACTER *character)
{

	if (enemies_noticed(character)) {

		return AI_STATE_COMBAT;
	}

	if (character->party != PARTY_PLAYER) {

		if (RANDOM_CHOICE(AI_SWITCH_TO_PATROL_PROBABILITY) ||
			is_standing_on_elevator(character)) {

			return AI_STATE_PATROL;
		}
	}

	ai_get_ready_action(character);

	return AI_STATE_GUARD;
}



/*
 * AI state: patrol
 */
static AI_STATE ai_state_patrol(CHARACTER *character)
{
	PATH_NODE *path;

	if (enemies_noticed(character)) {

		return AI_STATE_COMBAT;
	}

	if (RANDOM_CHOICE(AI_SWITCH_TO_GUARD_PROBABILITY)) {

		return AI_STATE_GUARD;
	}
	
	ai_get_ready(character);
	
	if (character->action_spent) {
	
		return AI_STATE_PATROL;
	}
	
	if (!ai_has_target(character)) {
		AREA_POINT target_point;

		random_sector(&target_point, SC_MOVE_TARGET_SAFE);

		ai_set_target(character, &target_point);
		
		action_do_nothing(character);

		return AI_STATE_PATROL;
	}
	
	if (ai_target_reached(character)) {
		
		ai_erase_target(character);
		
		action_do_nothing(character);

		return AI_STATE_PATROL;
	}
	
	path = find_patrol_path(character, &character->ai.target_point);
 
	if (path != NULL) {
		CHARACTER *bumped_into;
		
		bumped_into = character_at(&path->p);

		if (bumped_into != NULL &&
			bumped_into_unnoticed_enemy(character, bumped_into)) {

			return AI_STATE_PATROL;
		}

		if (action_move(character, &path->p)) {

			return AI_STATE_PATROL;
		}
	}
	
	ai_erase_target(character);

	action_do_nothing(character);

	return AI_STATE_PATROL;
}



/*
 * AI state: follow player controlled character
 */
static AI_STATE ai_state_follow_pcc(CHARACTER *character)
{
	PATH_NODE *path;
	SECTOR *sector;

	if (enemies_noticed(character)) {

		if (character->ai.tactic == TACTIC_PLAN_A ||
			ai_close_to_pcc(character)) {

			return AI_STATE_COMBAT;
		}
	}
	
	ai_set_target(character, &player_controlled_character()->location);

	if (character->ai.target_distance == 1) {
		
		ai_get_ready_action(character);
		
		return AI_STATE_FOLLOW_PCC;
	}

	path = find_run_path(character, &character->ai.target_point);

	if (path == NULL) {

		ai_get_ready_action(character);

		return AI_STATE_FOLLOW_PCC;
	}

	sector = sector_at(&path->p);

	if (sector->character != NULL) {
		
		if (bumped_into_unnoticed_enemy(character, 
			sector->character)) {

			return AI_STATE_FOLLOW_PCC;
		}

		if (can_push_past(character, sector->character)) {

			action_push_past(character, sector->character);

			return AI_STATE_FOLLOW_PCC;
		}

		path = find_run_around_character_path(character,
			&character->ai.target_point
		);

		if (path != NULL && 
			path_length(path) <= MAX_SHORT_PATH) {

			sector = sector_at(&path->p);
			
		} else {
		
			ai_get_ready_action(character);

			return AI_STATE_FOLLOW_PCC;
		}
	} 

	if (sector->object != NULL &&
		object_has_attribute(sector->object, OA_IMPASSABLE)) {

		ai_set_target(character, &path->p);

		return AI_STATE_ATTACK;
	}

	action_run(character, path, run_factor(character));

	return AI_STATE_FOLLOW_PCC;
}



/*
 * AI state: combat
 */
static AI_STATE ai_state_combat(CHARACTER *character)
{
	CHARACTER *nearest_enemy;

	if (character->ai.tactic == TACTIC_GET_READY) {

		ai_set_tactic(character, TACTIC_PLAN_A);
	}

	if (character->ai.tactic == TACTIC_STAY_CLOSE &&
		!ai_close_to_pcc(character)) {

		return AI_STATE_FOLLOW_PCC;
	}

	nearest_enemy = nearest_noticed_enemy(character);

	if (nearest_enemy != NULL) {

		ai_set_target(character,
			&nearest_enemy->location
		);

		if (character->perk[PK_PSYCHIC]) {

			return AI_STATE_PSYKER_COMBAT;
		}

		return AI_STATE_ATTACK;
	}

	if (!ai_has_target(character) ||
		character->party == PARTY_PLAYER) {

		return character->ai.default_state;
	}

	if (ai_target_reached(character)) {

		return character->ai.default_state;
	}

	return AI_STATE_COMBAT_RUN_TOWARDS_TARGET;
}



/*
 * AI state: combat run towards target
 */
static AI_STATE ai_state_combat_run_towards_target(CHARACTER *character)
{
	PATH_NODE *path;
	SECTOR *sector;

	path = find_run_path(character, &character->ai.target_point);
	
	if (path == NULL) {

		action_do_nothing(character);

		return character->ai.default_state;
	}

	sector = sector_at(&path->p);

	if (sector->character != NULL) {
		
		if (bumped_into_unnoticed_enemy(character, 
			sector->character)) {

			return AI_STATE_COMBAT;
		}

		if (can_push_past(character, sector->character)) {

			action_push_past(character, sector->character);

			return AI_STATE_COMBAT;
		}

		if (character->ai.target_distance == 1) {

			ai_get_ready_action(character);;

			return AI_STATE_COMBAT;
		}

		path = find_run_around_character_path(character,
			&character->ai.target_point
		);
		
		if (path != NULL && 
			path_length(path) <= MAX_SHORT_PATH) {

			sector = sector_at(&path->p);
			
		} else {
		
			ai_get_ready_action(character);

			return AI_STATE_COMBAT;
		}
	}
	
	if (sector->object != NULL &&
		object_has_attribute(sector->object, OA_IMPASSABLE)) {

		ai_set_target(character, &path->p);

		return AI_STATE_ATTACK;
	}
	
	action_run(character, path, run_factor(character));

	return AI_STATE_COMBAT;
}



/*
 * AI state: psyker - combat
 */
static AI_STATE ai_state_psyker_combat(CHARACTER *character)
{
	NOT_USED(character);

	if (RANDOM_CHOICE(AI_PSYKER_DEFENSIVE_ACTION_PROBABILITY)) {

		return AI_STATE_PSYKER_DEFENSIVE;
	}

	return AI_STATE_PSYKER_OFFENSIVE;
}



/*
 * AI state: psyker - defensive
 */
static AI_STATE ai_state_psyker_defensive(CHARACTER *character)
{

	if (character->psy_power[PSY_REGENERATE] &&
		character->injury > injury_max(character) / 2 &&
		ai_psy_power_sensible_choice(
			PSY_REGENERATE, &character->location)) {

		ai_set_target(character, &character->location);
		character->ai.chosen_power = PSY_REGENERATE;

	} else if (character->psy_power[PSY_PSYCHIC_SHIELD] &&
		ai_psy_power_sensible_choice(
			PSY_PSYCHIC_SHIELD, &character->location)) {

		ai_set_target(character, &character->location);
		character->ai.chosen_power = PSY_PSYCHIC_SHIELD;

	} else {

		return AI_STATE_ATTACK;
	}

	return AI_STATE_PSYKER_EVOCATION;
}



/*
 * AI state: psyker - offensive
 */
static AI_STATE ai_state_psyker_offensive(CHARACTER *character)
{

	if (character->ai.target_distance > MAX_LONG_RANGE &&
		character->ai.tactic == TACTIC_PLAN_A) {

		return AI_STATE_COMBAT_RUN_TOWARDS_TARGET;
	}

	if (character->psy_power[PSY_TERRIFY] &&
		ai_psy_power_sensible_choice(
			PSY_TERRIFY, &character->ai.target_point)) {

		character->ai.chosen_power = PSY_TERRIFY;

	} else if (character->psy_power[PSY_PSYCHIC_SHRIEK] &&
		ai_psy_power_sensible_choice(PSY_PSYCHIC_SHRIEK,
			&character->ai.target_point)) {

		character->ai.chosen_power = PSY_PSYCHIC_SHRIEK;

	} else if (character->psy_power[PSY_FIRESTORM] &&
		ai_psy_power_sensible_choice(PSY_FIRESTORM,
			&character->ai.target_point)) {

		character->ai.chosen_power = PSY_FIRESTORM;

	} else if (character->psy_power[PSY_FIREBALL] &&
		ai_psy_power_sensible_choice(PSY_FIREBALL,
			&character->ai.target_point)) {

		character->ai.chosen_power = PSY_FIREBALL;

	} else {

		return AI_STATE_ATTACK;
	}

	return AI_STATE_PSYKER_EVOCATION;
}



/*
 * AI state: psyker - evocation
 */
static AI_STATE ai_state_psyker_evocation(CHARACTER *character)
{

	if (psy_power_difficulty(character, character->ai.chosen_power) >=
		AI_MIN_DIFFICULTY_TO_CONCENTRATE &&
		character->concentration < AI_NORMAL_CONCENTRATION) {

		action_concentrate(character);

		return AI_STATE_COMBAT;
	}

	action_evoke_psy_power(character, character->ai.chosen_power);

	return AI_STATE_COMBAT;
}



/*
 * AI state: attack
 */
static AI_STATE ai_state_attack(CHARACTER *character)
{

	if (character->ai.target_distance > MAX_CLOSE_COMBAT_RANGE) {

		return AI_STATE_ATTACK_TARGET_DISTANT;
	}

	return AI_STATE_ATTACK_TARGET_CLOSE;
}



/*
 * AI state: attack - target close
 */
static AI_STATE ai_state_attack_target_close(CHARACTER *character)
{
	const OBJECT_DATA *weapon_data;

	if (character->weapon != NULL) {

		weapon_data = object_static_data(character->weapon);

		if (weapon_data->type == OTYPE_CLOSE_COMBAT_WEAPON) {

			return AI_STATE_ATTACK_STRIKE;
		}

		if (weapon_data->subtype == OSTYPE_PISTOL) {

			if (character->secondary_weapon == NULL) {

				return AI_STATE_ATTACK_SHOOT;
			}

			weapon_data = object_static_data(
				character->secondary_weapon
			);

			if (weapon_data->type ==
				OTYPE_CLOSE_COMBAT_WEAPON) {

				action_switch_weapons(character);

				return AI_STATE_COMBAT;
			}

			return AI_STATE_ATTACK_SHOOT;
		}
	}

	if (character->secondary_weapon == NULL) {

		if (character->weapon == NULL) {

			return AI_STATE_ATTACK_STRIKE;

		} else {

			action_switch_weapons(character);

			return AI_STATE_COMBAT;
		}
	}

	weapon_data = object_static_data(character->secondary_weapon);

	if (weapon_data->type == OTYPE_CLOSE_COMBAT_WEAPON ||
		weapon_data->subtype == OSTYPE_PISTOL) {

		action_switch_weapons(character);

		return AI_STATE_COMBAT;
	}

	if (character->weapon != NULL) {

		action_unequip_object(character, character->weapon);

		return AI_STATE_COMBAT;
	}

	return AI_STATE_ATTACK_STRIKE;
}



/*
 * AI state: attack - target distant
 */
static AI_STATE ai_state_attack_target_distant(CHARACTER *character)
{

	if (character->ai.target_distance > MAX_LONG_RANGE &&
		!character->perk[PK_MARKSMAN] &&
		character->ai.tactic == TACTIC_PLAN_A) {

		return AI_STATE_COMBAT_RUN_TOWARDS_TARGET;
	}

	if (character->ai.target_distance <= MAX_MEDIUM_RANGE &&
		ai_close_combat_focused(character) &&
		character->ai.tactic == TACTIC_PLAN_A) {

		return AI_STATE_COMBAT_RUN_TOWARDS_TARGET;
	}

	if (character->weapon != NULL &&
		object_static_data(character->weapon)->type ==
			OTYPE_RANGED_COMBAT_WEAPON) {

		return AI_STATE_ATTACK_SHOOT;
	}

	if (character->secondary_weapon == NULL) {

		if (character->ai.tactic == TACTIC_PLAN_A) {

			return AI_STATE_COMBAT_RUN_TOWARDS_TARGET;

		} else {

			action_do_nothing(character);

			return AI_STATE_COMBAT;
		}
	}

	if (object_static_data(character->secondary_weapon)->type ==
		OTYPE_RANGED_COMBAT_WEAPON) {

		action_switch_weapons(character);

		return AI_STATE_COMBAT;
	}

	if (character->ai.tactic == TACTIC_PLAN_A) {

		return AI_STATE_COMBAT_RUN_TOWARDS_TARGET;
	}

	action_do_nothing(character);

	return AI_STATE_COMBAT;
}



/*
 * AI state: attack strike
 */
static AI_STATE ai_state_attack_strike(CHARACTER *character)
{

	action_strike(character, character->weapon,
		&character->ai.target_point
	);

	return AI_STATE_COMBAT;
}



/*
 * AI state: attack shoot
 */
static AI_STATE ai_state_attack_shoot(CHARACTER *character)
{
	FIRING_DATA firing_data;
	OBJECT *weapon;
	const OBJECT_DATA *weapon_data;
	const FIRING_MODE_DATA *firing_mode;

	weapon = character->weapon;
	weapon_data = object_static_data(weapon);

	firing_mode = &weapon_data->firing_mode;

	if (!weapon->functional) {

		action_unjam_weapon(character);

		return AI_STATE_COMBAT;
	}

	if (weapon->charge == 0) {

		if (weapon_data->attribute[OA_AUTOMATIC_RECHARGE]) {

			action_do_nothing(character);

		} else {

			action_reload_weapon(character);
		}

		return AI_STATE_COMBAT;
	}

	if (firing_mode->has[FMODE_A]) {

		firing_data.firing_mode = FMODE_A;
		firing_data.n_shots = firing_mode->a_shots;
		firing_data.spread = random_int(A_MODE_SPREAD_MIN,
			A_MODE_SPREAD_MAX);

	} else if (firing_mode->has[FMODE_SA]) {

		firing_data.firing_mode = FMODE_SA;
		firing_data.n_shots = random_int(
			firing_mode->min_sa_shots,
			firing_mode->max_sa_shots
		);

	} else {

		firing_data.firing_mode = FMODE_S;
		firing_data.n_shots = 1;
	}

	action_shoot(character, weapon, &firing_data,
		&character->ai.target_point
	);

	return AI_STATE_COMBAT;
}



/*
 * AI state: recover
 */
static AI_STATE ai_state_recover(CHARACTER *character)
{

	if (enemies_noticed(character)) {

		return AI_STATE_COMBAT;
	}

	action_recover(character);

	return character->ai.default_state;
}



/*
 * AI state: flee
 */
static AI_STATE ai_state_flee(CHARACTER *character)
{
	CHARACTER *nearest_enemy;
	AREA_POINT target_point;
	AREA_DISTANCE i;

	nearest_enemy = nearest_noticed_enemy(character);

	if (nearest_enemy == NULL) {

		action_do_nothing(character);

		return AI_STATE_FLEE;
	}

	target_point = character->location;

	character->is_running = true;

	for (i = run_factor(character); i > 0; --i) {

		coordinate_step_away(&target_point, 
			&nearest_enemy->location
		);

		if (out_of_area_bounds(&target_point)) {

			break;

		} else if (!sector_is_class(&target_point,
			SC_MOVE_TARGET)) {

			break;
		}

		if (!action_move(character, &target_point)) {

			break;
		}
	}

	character->is_running = false;

	if (!character->action_spent) {

		action_do_nothing(character);
	}

	return AI_STATE_FLEE;
}



/*
 * AI state: player control
 */
static AI_STATE ai_state_player_control(CHARACTER *character)
{
	COMMAND entered_command;

	use_game_screen_command_bar();
	render_game_screen();
	update_screen();

	entered_command = command_bar_get_command();

	clear_message_buffer(SID_GAME);

	handle_command(character, entered_command);

	return character->ai.default_state;
}



/*
 * updates the target distance
 */
static void ai_update_target_distance(CHARACTER *character)
{
	
	character->ai.target_distance = area_distance(
		&character->location, &character->ai.target_point
	);
}



/*
 * makes a character attempt to 'fix' himself
 * e.g. taking an antidote when poisoned
 */
static void ai_fix_self(CHARACTER *character)
{

	if (character_has_flag(character, CF_POISONED)) {

		if (ai_use_drug(character, "Antidote")) {

			return;
		}
	}

	if (character->perk[PK_STEALTH] &&
		!character_unnoticed(character)) {

		if (action_stealth(character)) {

			return;
		}
	}

	if (character_has_flag(character, CF_BROKEN)) {

		if (ai_use_drug(character, "Stoic")) {

			return;
		}
	}
}



/*
 * makes a character take the named drug
 * if it is present in his inventory
 */
static bool ai_use_drug(CHARACTER *character, const char *drug_name)
{
	OBJECT *drug;

	drug = inventory_find_object(character, drug_name);

	if (drug == NULL) {

		return false;
	}

	return action_use_drug(character, drug);
}



/*
 * returns true if the character's AI is close combat focused
 */
static bool ai_close_combat_focused(const CHARACTER *character)
{

	if (character->stat[S_CC].current >=
		character->stat[S_RC].current + 20) {

		return true;
	}

	return false;
}



/*
 * returns true if the passed psychic power is a 'sensible choice'
 */
static bool ai_psy_power_sensible_choice(PSY_POWER power,
	const AREA_POINT *target_point
)
{

	if (!psy_power_valid_target(power, target_point)) {

		return false;
	}

	if (psy_power_in_effect(power, target_point)) {

		return false;
	}

	if (psy_power_target_is_immune(power, target_point)) {

		return false;
	}

	return true;
}


/*
 * returns true if the passed character is close to
 * the player controlled character
 */
static bool ai_close_to_pcc(const CHARACTER *character)
{
	const AREA_POINT *p1;
	const AREA_POINT *p2;

	p1 = &character->location;
	p2 = &player_controlled_character()->location;

	if (area_distance(p1, p2) > 2) {

		return false;
	}

	return true;
}



/*
 * sets the target of the AI
 */
static void ai_erase_target(CHARACTER *character)
{

	character->ai.target_point = *area_point_nil();

	character->ai.target_distance = AREA_DISTANCE_NIL;
}



/*
 * returns true if the passed character has reached his target
 */
static bool ai_target_reached(const CHARACTER *character)
{

	if (area_points_equal(&character->location,
		&character->ai.target_point)) {

		return true;
	}

	return false;
}



/*
 * returns true if the AI has a target
 */
static bool ai_has_target(const CHARACTER *character)
{

	if (character->ai.target_distance == AREA_DISTANCE_NIL) {
		
		return false;
	}

	return true;
}



/*
 * exactly like ai_get_ready() except that this function
 * will make sure that an action is executed 
 * if there is nothing to do it will just do nothing
 */
static void ai_get_ready_action(CHARACTER *character)
{
	
	ai_get_ready(character);
	
	if (!character->action_spent) {

		action_do_nothing(character);
	}	
}



/*
 * makes the passed character ready himself for combat
 */
static void ai_get_ready(CHARACTER *character)
{
	
	if (character->weapon != NULL) {

		ai_ready_weapon(character);
		
		if (character->action_spent) {

			return;
		}
	}
	
	if (character_must_recover(character)) {

		action_recover(character);

		return;
	}

	if (character->ai.tactic == TACTIC_GET_READY &&
		character->perk[PK_PSYCHIC]) {
		
		ai_get_ready_psyker(character);
	}
}



/*
 * makes the passed character ready his weapon
 */
static void ai_ready_weapon(CHARACTER *character)
{
	OBJECT *weapon;
	const OBJECT_DATA *weapon_data;

	weapon = character->weapon;

	weapon_data = object_static_data(weapon);

	if (!weapon->functional) {

		action_unjam_weapon(character);

		return;
	}
	
	if (weapon->charge < weapon_data->charge_max &&
		!weapon_data->attribute[OA_AUTOMATIC_RECHARGE]) {

		action_reload_weapon(character);

		return;
	}
}



/*
 * makes the passed character ready himself for evoking psychic powers
 */
static void ai_get_ready_psyker(CHARACTER *character)
{

	if (character->concentration == MAX_CONCENTRATION_BONUS) {

		return;
	}

	action_concentrate(character);
}



/*
 * updates the awareness data of the passed character
 */
static void ai_update_awareness_data(CHARACTER *character)
{
	CHARACTER *target;
	LIST_NODE *node;
	LIST *noticed_enemies;

	noticed_enemies = character->ai.noticed_enemies;

	list_clear(noticed_enemies);

	for (node = area_character_list()->head;
		node != NULL;
		node = node->next) {

		target = (CHARACTER *)node->data;

		if (!hostility_between(character, target)) {

			continue;
		}

		if (!los(&character->location, &target->location)) {

			continue;
		}
		
		if (character_unnoticed(target)) {

			notice_check(target, true);
		}

		if (!character_unnoticed(target)) {

			list_add(noticed_enemies, target);

			continue;
		}
	}
}



