/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Character
 * Description: -
 */

#define Uses_Ui
#define Uses_CharacterAdvancement
#define Uses_Inventory
#define Uses_Util
#define Uses_Object
#define Uses_Career
#define Uses_Stats
#define Uses_Perks
#define Uses_Psychic
#define Uses_Party
#define Uses_Dice
#define Uses_ProgramManager
#define Uses_Event
#define Uses_DataFile
#define Uses_Effect
/* hack - fix later */
#define Uses_DynamicMessage
#define Uses_Faction
#define Uses_Round
#define Uses_Area
#include "mheader.h"

#include "charact.h"


typedef struct character_flag_data_struct	CHARACTER_FLAG_DATA;


/*
 * character flag data structure
 */
struct character_flag_data_struct {

	const char *	name;

	const COLOR	color;

};


static void	character_representation_init(CHARACTER *);
static void	character_state_init(CHARACTER *);
static void	character_stats_init(CHARACTER *);
static void	character_flags_init(CHARACTER *);
static void	character_perks_init(CHARACTER *);
static void	character_psy_powers_init(CHARACTER *);
static void	character_advancement_init(CHARACTER *);
static void	character_party_data_init(CHARACTER *);
static void	character_inventory_init(CHARACTER *);
static void	character_equipment_init(CHARACTER *);
static void	character_ai_init(CHARACTER *);
static void	character_serialization_init(CHARACTER *);


static void	character_screen_setup_command_bar(const CHARACTER *);

static void	character_screen_basics(const CHARACTER *);



static const CHARACTER_FLAG_DATA CharacterFlag[MAX_CHARACTER_FLAGS] = {

	{"Unique",		C_UNIQUE},
	{"Daemonic",		C_DAEMONIC},
	{"Machine",		C_MACHINE},
	{"Boss",		C_BOSS},

	{"Stunned",		C_STUNNED},
	{"Poisoned",		C_POISONED},
	{"Diseased",		C_DISEASED},
	{"Broken",		C_BROKEN},
	{"Psychic overload",	C_PSYCHIC_OVERLOAD},
	{"Busy",		C_TEXT},

	{"Warp strength",	C_PSY_POWER},
	{"Psychic shield",	C_PSY_POWER},

	{"Flash",		C_DRUG},
	{"Stoic",		C_DRUG},
	{"Nomat",		C_DRUG},
	{"Zen",			C_DRUG},
	{"Bull",		C_DRUG},

	{"Unnoticed",		C_UNNOTICED},

	{"Cannot speak",	C_CANNOT_SPEAK}

};



static CHARACTER *	PlayerCharacter = NULL;

static CHARACTER *	PlayerControlledCharacter = NULL;



/*
 * allocates memory for a character
 */
CHARACTER * character_new(void)
{
	CHARACTER *character;

	character = checked_malloc(sizeof *character);

	character_init(character);

	character->inventory = list_new();
	character->ai.noticed_enemies = list_new();

	return character;
}



/*
 * destroys a character
 */
void character_destroy(void *data)
{
	CHARACTER *character;

	character = (CHARACTER *)data;

	if (character->list_node != NULL) {
		LIST *character_events;

		character_events = list_new();

		remove_character(character, character_events);

		list_destroy_with(character_events, event_destroy);
	}

	if (character == PlayerControlledCharacter) {
		PlayerControlledCharacter = NULL;
	}

	if (character == PlayerCharacter) {
		PlayerCharacter = NULL;
	}

	list_destroy_with(character->inventory, object_destroy);

	list_destroy(character->ai.noticed_enemies);

	free(character);
}



/*
 * initializes a character with default values
 */
void character_init(CHARACTER *character)
{

	character_representation_init(character);

	character_state_init(character);

	character_stats_init(character);

	character_flags_init(character);

	character_perks_init(character);

	character_psy_powers_init(character);

	character_advancement_init(character);

	character_party_data_init(character);

	character_inventory_init(character);

	character_equipment_init(character);

	character_ai_init(character);

	character_serialization_init(character);
}



/*
 * produces an identical copy of a character
 */
CHARACTER * character_clone(const CHARACTER *character)
{
	CHARACTER *clone;
	LIST_NODE *node;

	clone = checked_malloc(sizeof *clone);
	*clone = *character;

	clone->inventory = list_new();

	for (node = character->inventory->head;
		node != NULL;
		node = node->next) {
		OBJECT *object;
		OBJECT *cloned_object;

		object = (OBJECT *)node->data;
		cloned_object = object_clone(object);

		inventory_add(clone, cloned_object);

		if (object == character->weapon) {
			
			clone->weapon = cloned_object;
			
		} else if (object == character->secondary_weapon) {
			
			clone->secondary_weapon = cloned_object;
			
		} else if (object == character->armour) {
			
			clone->armour = cloned_object;
			
		} else if (object == character->jump_pack) {
			
			clone->jump_pack = cloned_object;
		}
	}

	clone->ai.noticed_enemies = list_new();

	return clone;
}



/*
 * makes the passed character the new player character
 */
void set_player_character(CHARACTER *character)
{

	PlayerCharacter = character;
}



/*
 * returns the player character
 */
CHARACTER * player_character(void)
{

	return PlayerCharacter;
}



/*
 * makes the passed character the new player controlled character
 */
void set_player_controlled_character(CHARACTER *character)
{

	if (PlayerControlledCharacter != NULL) {
		PlayerControlledCharacter->controller = CC_AI;
		PlayerControlledCharacter->ai.default_state = AI_STATE_FOLLOW_PCC;
		PlayerControlledCharacter->ai.state = AI_STATE_FOLLOW_PCC;
		PlayerControlledCharacter->symbol = 'H';
	}

	PlayerControlledCharacter = character;

	if (PlayerControlledCharacter == NULL) return;

	PlayerControlledCharacter->controller = CC_PLAYER;
	PlayerControlledCharacter->ai.default_state = AI_STATE_PLAYER_CONTROL;
	PlayerControlledCharacter->ai.state = AI_STATE_PLAYER_CONTROL;
	PlayerControlledCharacter->symbol = '@';
}



/*
 * returns the player controlled character
 */
CHARACTER * player_controlled_character(void)
{

	return PlayerControlledCharacter;
}



/*
 * sets a character flag
 */
void character_set_flag(CHARACTER *character, CHARACTER_FLAG flag)
{

	character->flag[flag] = true;
}



/*
 * removes a character flag
 */
void character_remove_flag(CHARACTER *character, CHARACTER_FLAG flag)
{

	character->flag[flag] = false;
}



/*
 * returns true if the passed character has the passed flag
 */
bool character_has_flag(const CHARACTER *character, CHARACTER_FLAG flag)
{

	return character->flag[flag];
}



/*
 * returns the description of a character
 */
char * character_description(char *description, const CHARACTER *character)
{

	return data_file_character_description(description, character);
}



/*
 * the character screen
 */
void character_screen(CHARACTER *character)
{
	COMMAND entered_command;

	do {

		character_screen_setup_command_bar(character);
		character_screen_basics(character);

		entered_command = command_bar_get_command();

		if (entered_command == CM_PSY_POWERS) {

			psy_powers_screen(character);

		} else if (entered_command == CM_CAREER) {

			career_screen(get_career_pointer(character->career));

		} else if (entered_command == CM_ADVANCEMENT) {

			character_advancement_screen(character);

		} else if (entered_command == CM_RENAME) {

			name_character(character);
		}

	} while (entered_command != CM_EXIT);


}



/*
 * the name character screen
 */
void name_character(CHARACTER *character)
{
	GET_TEXT_DIALOG dialog;

	sprintf(dialog.prompt,
		"Name (limit: %d characters): ",
		CHARACTER_NAME_SIZE - 1
	);
	dialog.prompt_color = C_FIELD_NAME;
	dialog.max_length = CHARACTER_NAME_SIZE - 1;

	command_bar_set(1, CM_OK);

	render_name_character_screen();
	update_screen();

	enter_name_dialog(&dialog);

	strcpy(character->name, dialog.entered_text);

	if (strlen(character->name) < 2) strcpy(character->name, "Nobody");

	if (!character_has_flag(character, CF_UNIQUE)) {

		character_set_flag(character, CF_UNIQUE);
	}
}



/*
 * returns the power rating of a character
 */
POWER_RATING character_power_rating(const CHARACTER *character)
{
	int i;
	POWER_RATING power_rating;
	LIST_NODE *node;

	power_rating = 0;

	for (i = 0; i < MAX_STATS; i++) {

		if (i == S_SG) {
			
			continue;
		}

		if (i == S_FR && !character->perk[PK_PSYCHIC]) {

			continue;
		}

		if (i == S_LD &&
			character_has_flag(character, CF_MACHINE)) {

			power_rating += 100;

			continue;
		}

		power_rating += character->stat[i].total;
	}

	for (i = 0; i < MAX_PERKS; i++) {
		
		if (character->perk[i]) {

			power_rating += 10;
		}
	}

	for (i = 0; i < MAX_PSY_POWERS; i++) {
		
		if (character->psy_power[i]) {
			
			power_rating += 10;
		}
	}

	for (node = character->inventory->head;
		node != NULL;
		node = node->next) {
		const OBJECT_DATA *object_data;

		object_data = object_static_data(node->data);

		if (object_data->type == OTYPE_MONEY ||
			object_data->value == VALUE_NIL) {

			continue;
		}

		power_rating += object_data->value / 100;
	}

	return power_rating;
}



/*
 * prints a character dump
 */
void character_dump_print(const CHARACTER *character, FILE *out_file)
{
	LIST_NODE *node;
	int i;

	fprintf(out_file, "Name        : %s\n", character->name);

	fprintf(out_file, "Career      : ");
	if (character->career == CAREER_NIL) {
		fprintf(out_file, "-\n");
	} else {
		fprintf(out_file, "%s\n",
			get_career_pointer(character->career)->name
		);
	}

	fprintf(out_file, "Power Rating: %d\n", character_power_rating(character));

	fprintf(out_file, "\n");

	for (i = 0; i < MAX_STATS; i++) {
		fprintf(out_file, "%-15s: %d\n", stat_long_name(i),
			character->stat[i].current
		);
	}

	fprintf(out_file, "\n[Perks]\n");
	for (i = 0; i < MAX_PERKS; i++) {
		if (character->perk[i]) {
			fprintf(out_file, "%s\n", perk_name(i));
		}
	}

	fprintf(out_file, "\n[Status]\n");
	if (character->injury > 0) {
		
		fprintf(out_file, "Injured (%d/%d)\n",
			character->injury,
			injury_max(character)
		);
	}

	for (i = 0; i < MAX_CHARACTER_FLAGS; i++) {

		if (!character_has_flag(character, i)) {
			continue;
		}

		if (i == CF_UNIQUE) continue;

		fprintf(out_file, "%s\n", character_flag_name(i));
	}

	fprintf(out_file, "\n[Psychic Powers]\n");
	for (i = 0; i < MAX_PSY_POWERS; i++) {

		if (!character->psy_power[i]) continue;
		fprintf(out_file, "%s\n", psy_power_name(i));
	}

	fprintf(out_file, "\n[Inventory]\n");
	for (node = character->inventory->head;
		node != NULL;
		node = node->next) {

		inventory_object_print(character, (OBJECT *)node->data,
			out_file
		);
	}

	fprintf(out_file, "\n\n");

}



/*
 * the character dump screen
 */
void character_dump_screen(void)
{
	party_dump_print();

	command_bar_set(1, CM_EXIT);
	render_character_dump_screen();
	update_screen();

	command_bar_get_command();
}




/*
 * effect of the recover action
 */
void character_recover(CHARACTER *character)
{
	int test_value;

	test_value = character->stat[S_TN].current;
	
	if (character->perk[PK_FIRST_AID]) {

		test_value += FIRST_AID_RECOVERY_BONUS;
	}

	if (character_has_flag(character, CF_DISEASED) &&
		d100_test_passed(test_value + 
			DISEASED_RECOVERY_MODIFIER)) {

		event_execute(
			effect_termination_event(character, ET_DISEASED)
		);
			
		return;
	}

	if (character_has_flag(character, CF_PSYCHIC_OVERLOAD) &&
		d100_test_passed(character->stat[S_FR].current +
			PSYCHIC_OVERLOAD_RECOVERY_MODIFIER)) {

		event_execute(
			effect_termination_event(character, 
				ET_PSYCHIC_OVERLOAD)
		);

		return;
	}

	if (character->injury > 0) {

		if (d100_test_passed(test_value)) {
		
			character_recover_injuries(character);
		}
	}
}



/*
 * recover injuries effect
 */
void character_recover_injuries(CHARACTER *character)
{
	int bonus;

	bonus = stat_bonus(character, S_TN);

	character->injury -= dice(1, 3) + bonus;
	
	if (character->injury < 0) {
		
		character->injury = 0;
	}
}



/*
 * returns true if the character must recover to regain full strength
 */
bool character_must_recover(const CHARACTER *character)
{

	if (character->injury > 0 ||
		character_has_flag(character, CF_DISEASED) ||
		character_has_flag(character, CF_PSYCHIC_OVERLOAD)) {

		return true;
	}

	return false;
}



/*
 * returns the maximal number of injuries the passed character can survive
 */
INJURY injury_max(const CHARACTER *character)
{
	INJURY maximum;
	
	maximum = character->stat[S_TN].total;
	
	if (character->perk[PK_DIE_HARD]) {
	
		maximum += percent(DIE_HARD_MAX_INJURY_BONUS, maximum);
	}
	
	return maximum;
}



/*
 * returns the name of a character flag
 */
const char * character_flag_name(CHARACTER_FLAG flag)
{

	return CharacterFlag[flag].name;
}



/*
 * returns the color of a character flag
 */
COLOR character_flag_color(CHARACTER_FLAG flag)
{

	return CharacterFlag[flag].color;
}



/*
 * name -> character flag
 */
CHARACTER_FLAG name_to_character_flag(const char *name)
{
	int i;

	for (i = 0; i < MAX_CHARACTER_FLAGS; i++) {

		if (STR_EQ(name, CharacterFlag[i].name)) {
			return i;
		}

	}

	die("*** CORE ERROR *** invalid character flag (%s)", name);

	/* NEVER REACHED */
	return 0;
}



/*
 * initializes a character's representation data
 */
static void character_representation_init(CHARACTER *character)
{

	CLEAR_STRING(character->name);

	character->symbol = 'H';

	character->color = C_DEFAULT;

	character->gender = GENDER_NEUTER;
}



/*
 * initializes a character's state
 */
static void character_state_init(CHARACTER *character)
{

	character->type = CT_NPC;
	character->controller = CC_AI;
	character->location = *area_point_nil();

	character->is_hostile = true;

	character->injury = 0;

	character->bloody_feet = 0;

	character->concentration = 0;

	character->is_running = false;

	character->action_spent = false;
}



/*
 * initializes a character's stats
 */
static void character_stats_init(CHARACTER *character)
{
	int i;

	for (i = 0; i < MAX_STATS; i++) {
		CHARACTER_STAT *stat;

		stat = &(character->stat[i]);

		stat->current = stat->total = 0;

		stat->advance = 0;
	}
}



/*
 * initializes a character's status flags
 */
static void character_flags_init(CHARACTER *character)
{
	bool *flag;
	int i;

	flag = character->flag;

	for (i = 0; i < MAX_CHARACTER_FLAGS; i++) {

		flag[i] = false;
	}
}



/*
 * initializes a character's perks
 */
static void character_perks_init(CHARACTER *character)
{
	bool *perk;
	int i;

	perk = character->perk;

	for (i = 0; i < MAX_PERKS; i++) {

		perk[i] = false;
	}
}



/*
 * initializes a character's psychic powers
 */
static void character_psy_powers_init(CHARACTER *character)
{
	bool *power;
	int i;

	power = character->psy_power;

	for (i = 0; i < MAX_PSY_POWERS; i++) {

		power[i] = false;
	}
}



/*
 * initializes a character's advancement data
 */
static void character_advancement_init(CHARACTER *character)
{

	character->career = CAREER_NIL;

	character->ep = 0;
}



/*
 * initializes a character's party data
 */
static void character_party_data_init(CHARACTER *character)
{

	character->party = PARTY_NIL;

	character->recruitable = true;
}



/*
 * initializes a character's inventory
 */
static void character_inventory_init(CHARACTER *character)
{

	character->inventory = NULL;
}



/*
 * initializes a character's equipment
 */
static void character_equipment_init(CHARACTER *character)
{

	character->weapon = NULL;
	character->secondary_weapon = NULL;
	character->armour = NULL;
	character->jump_pack = NULL;

	character->armour_rating.current =
		character->armour_rating.total = 0;
}


/*
 * initializes a character's AI data
 */
static void character_ai_init(CHARACTER *character)
{
	AI_DATA *ai;

	ai = &(character->ai);

	ai->state = ai->default_state = ai->original_state =
		AI_STATE_GUARD;

	ai->tactic = TACTIC_PLAN_A;

	ai->target_point = *area_point_nil();

	ai->target_distance = AREA_DISTANCE_NIL;

	ai->noticed_enemies = NULL;

	ai->chosen_power = PSY_NIL;
}



/*
 * initializes a character's serialization data
 */
static void character_serialization_init(CHARACTER *character)
{

	character->list_node = NULL;

	character->list_index = LIST_NODE_INDEX_NIL;
}



/*
 * sets up the character screen command bar
 */
static void character_screen_setup_command_bar(const CHARACTER *character)
{
	CAREER *career;

	career = get_career_pointer(character->career);

	command_bar_clear();

	if (n_psy_powers(character) > 0) {
		
		command_bar_add(1, CM_PSY_POWERS);
	}

	if (career != NULL) {

		command_bar_add(1, CM_CAREER);
	}

	if (character->party == PARTY_PLAYER) {
		
		command_bar_add(2, CM_ADVANCEMENT, CM_RENAME);
	}

	command_bar_add(1, CM_EXIT);
}



/*
 * the character screen basics
 */
static void character_screen_basics(const CHARACTER *character)
{
	render_character_screen_basics(character);

	update_screen();
}




